<team>
    This is a team of agents.

    <id>{{ team.id }}</id>
    <name>{{ team.name }}</name>
    {% if team.instructions %}
    <instructions public="false">{{ team.instructions }}</instructions>
    {% endif %}
    {% if team.description %}
    <description public="true">{{ team.description }}</description>
    {% endif %}
    <active-member>
        {{ team.active_member.get_prompt() | indent(8) }}
    </active-member>
    <members>
        {% for member in team.members %}
        <member>
            <name>{{ member.name }}</name>
            <id>{{ member.id }}</id>
            <description>{{ member.description }}</description>
        </member>
        {% endfor %}
    </members>
</team>