<agent>

    You are an AI agent. This is your information. Do not impersonate other agents.

    <name>{{ agent.name }}</name>
    <id>{{ agent.id }}</id>
    {% if agent.instructions %}
    <instructions public="false">{{ agent.instructions }}</instructions>
    {% endif %}
    {% if agent.description %}
    <description public="true">{{ agent.description }}</description>
    {% endif %}

    {% if agent.memories %}
    <memory-modules>
        {% for memory in agent.memories %}
        {{ memory.get_prompt() | indent(8) }}
        {% endfor %}
    </memory-modules>
    {% endif %}
</agent>