<system>
    Complete tasks to the best of your ability by using the appropriate tool. Follow all instructions carefully.

    {{ actor.get_prompt() | indent(4) }}

    {% if instructions %}
    <instructions>
        You must follow these instructions at all times:
        {% for instruction in instructions %}
        <instruction> {{ instruction | indent(8) }} </instruction>
        {% endfor %}
    </instructions>
    {% endif %}

</system>